@@ -19,20 +19,30 @@
  License along with GimpLensfun. If not, see
  http://www.gnu.org/licenses/.
 
- */
+###################################################################
+CHANGES:
+
+22.10.2011 - OpenMP parallelization
+20.10.2011 - Windows compatibility
+
+*/
 
 #include <stdio.h>
 #include <math.h>
-#include <iostream>
-#include <algorithm>
 #include <string>
 #include <vector>
 
 #include <lensfun.h>
 #include <libgimp/gimp.h>
 #include <libgimp/gimpui.h>
-#include <exiv2/image.hpp>
-#include <exiv2/exif.hpp>
+
+#ifdef _WIN32
+	#include <image.hpp>
+	#include <exif.hpp>
+#else
+	#include <exiv2/image.hpp>
+	#include <exiv2/exif.hpp>
+#endif
 
 using namespace std;
 
@@ -89,7 +99,6 @@ typedef struct
     std::string Camera;
     std::string CamMaker;
     std::string Lens;
-    //const char *LensMaker;
     float Scale;
     float Crop;
     float Focal;
@@ -433,15 +442,15 @@ modify_changed( GtkCheckButton *togglebutn,
 {
     sLensfunParameters.ModifyFlags = 0;
     if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(CorrDistortion)) 
-        && gtk_widget_get_sensitive(CorrDistortion))
+        && GTK_WIDGET_SENSITIVE(CorrDistortion))
         sLensfunParameters.ModifyFlags |= LF_MODIFY_DISTORTION;
     
     if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(CorrTCA)) 
-        && gtk_widget_get_sensitive(CorrTCA))
+        && GTK_WIDGET_SENSITIVE(CorrTCA))
         sLensfunParameters.ModifyFlags |= LF_MODIFY_TCA;
     
     if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(CorrVignetting)) 
-        && gtk_widget_get_sensitive(CorrDistortion))
+        && GTK_WIDGET_SENSITIVE(CorrDistortion))
         sLensfunParameters.ModifyFlags |= LF_MODIFY_VIGNETTING;
 }
 //####################################################################
@@ -698,7 +707,7 @@ inline int InterpolateNearest(guchar *ImgBuffer, gint imgwidth, gint channels, f
 //####################################################################
 // Processing
 static void process_image (GimpDrawable *drawable) {
-    gint         i, j, channels;
+    gint         channels;
     gint         x1, y1, x2, y2, imgwidth, imgheight;
 
     GimpPixelRgn rgn_in, rgn_out;
@@ -729,8 +738,6 @@ static void process_image (GimpDrawable *drawable) {
     //Init input and output buffer
     ImgBuffer = g_new (guchar, channels * (imgwidth) * (imgheight));
     ImgBufferOut = g_new (guchar, channels * imgwidth * imgheight);
-    // buffer containing undistorted coordinates for one row
-    float *UndistCoord = g_new (float, imgwidth*2*channels);
     
     // Copy pixel data from GIMP to internal buffer
     gimp_pixel_rgn_get_rect (&rgn_in, ImgBuffer, x1, y1, imgwidth, imgheight);
@@ -741,10 +748,6 @@ static void process_image (GimpDrawable *drawable) {
 
     const lfCamera **cameras = ldb->FindCamerasExt (sLensfunParameters.CamMaker.c_str(), sLensfunParameters.Camera.c_str());
     sLensfunParameters.Crop = cameras[0]->CropFactor;
-    
-    //sLensfunParameters.Scale = 0.95;
-    
-    //PrintCameras(cameras);
 
     const lfLens **lenses = ldb->FindLenses (cameras[0], NULL, sLensfunParameters.Lens.c_str());
 
@@ -753,44 +756,66 @@ static void process_image (GimpDrawable *drawable) {
     printf("Focal Length: %f\n", sLensfunParameters.Focal);
     printf("F-Stop: %f\n", sLensfunParameters.Aperture);
     printf("Crop Factor: %f\n", sLensfunParameters.Crop);
-    printf("Scale: %f\n", sLensfunParameters.Scale);
+    printf("Scale: %f\n", sLensfunParameters.Scale); 
     
-    //init lensfun modifier
-    lfModifier *mod = lfModifier::Create (lenses[0], sLensfunParameters.Crop, imgwidth, imgheight);
-    mod->Initialize (  lenses[0], LF_PF_U8, sLensfunParameters.Focal,
-                       sLensfunParameters.Aperture, sLensfunParameters.Distance, sLensfunParameters.Scale, sLensfunParameters.TargetGeom,
-                       sLensfunParameters.ModifyFlags, sLensfunParameters.Inverse);
-
-    //printf("Scale factor: %f \n", mod->GetAutoScale(false));
-    
-    //main loop for processing, iterate through rows
-    for (i = 0; i < imgheight; i++)
+    int iRowCount = 0;    
+    #pragma omp parallel
     {
-        mod->ApplyColorModification( &ImgBuffer[(channels*imgwidth*i)], 
-                                     0, i, imgwidth, 1, 
-                                     LF_CR_3(RED, GREEN, BLUE),
-                                     channels*imgwidth);
+        // buffer containing undistorted coordinates for one row        
+        float *UndistCoord = g_new (float, imgwidth*2*channels);
+
+        //init lensfun modifier
+        lfModifier *mod = lfModifier::Create (lenses[0], sLensfunParameters.Crop, imgwidth, imgheight);
+        mod->Initialize (  lenses[0], LF_PF_U8, sLensfunParameters.Focal,
+                        sLensfunParameters.Aperture, sLensfunParameters.Distance, sLensfunParameters.Scale, sLensfunParameters.TargetGeom,
+                        sLensfunParameters.ModifyFlags, sLensfunParameters.Inverse);
         
-        mod->ApplySubpixelGeometryDistortion (0, i, imgwidth, 1, UndistCoord);
-        float* CurrCoord = UndistCoord;
-        
-        //iterate through subpixels in one row
-        for (j = 0; j < imgwidth*channels; j += channels)
+        //main loop for processing, iterate through rows          
+        #pragma omp for      
+        for (int i = 0; i < imgheight; i++)
         {
-            if ((CurrCoord [0]>0) && (ceil(CurrCoord [0])<imgwidth) && (CurrCoord [1]>0) && (ceil(CurrCoord [1])<imgheight))  {
-                ImgBufferOut[(channels*imgwidth*i) + (j)  ] = InterpolateLinear(ImgBuffer, imgwidth, channels, CurrCoord [0], CurrCoord [1], 0);
-                ImgBufferOut[(channels*imgwidth*i) + (j)+1] = InterpolateLinear(ImgBuffer, imgwidth, channels, CurrCoord [0], CurrCoord [1], 1);
-                ImgBufferOut[(channels*imgwidth*i) + (j)+2] = InterpolateLinear(ImgBuffer, imgwidth, channels, CurrCoord [0], CurrCoord [1], 2);
-              } else {
-                ImgBufferOut[(channels*imgwidth*i) + (j)  ] = 0;
-                ImgBufferOut[(channels*imgwidth*i) + (j)+1] = 0;
-                ImgBufferOut[(channels*imgwidth*i) + (j)+2] = 0;
+            mod->ApplyColorModification( &ImgBuffer[(channels*imgwidth*i)], 
+                                        0, i, imgwidth, 1, 
+                                        LF_CR_3(RED, GREEN, BLUE),
+                                        channels*imgwidth);
+            
+            mod->ApplySubpixelGeometryDistortion (0, i, imgwidth, 1, UndistCoord);
+            
+            float*  CurrCoord = UndistCoord;
+            int     OutputBufferCoord = channels*imgwidth*i;	   
+            //iterate through subpixels in one row
+            for (int j = 0; j < imgwidth*channels; j += channels)
+            {
+                if ((CurrCoord [0]>0) && (ceil(CurrCoord [0])<imgwidth) && (CurrCoord [1]>0) && (ceil(CurrCoord [1])<imgheight))  {
+                    ImgBufferOut[OutputBufferCoord] = InterpolateLinear(ImgBuffer, imgwidth, channels, CurrCoord [0], CurrCoord [1], 0);
+                OutputBufferCoord++;
+                    ImgBufferOut[OutputBufferCoord] = InterpolateLinear(ImgBuffer, imgwidth, channels, CurrCoord [0], CurrCoord [1], 1);
+                OutputBufferCoord++;
+                    ImgBufferOut[OutputBufferCoord] = InterpolateLinear(ImgBuffer, imgwidth, channels, CurrCoord [0], CurrCoord [1], 2);
+                OutputBufferCoord++;
+                } else {
+                    ImgBufferOut[OutputBufferCoord] = 0;
+                OutputBufferCoord++;
+                    ImgBufferOut[OutputBufferCoord] = 0;
+                OutputBufferCoord++;
+                    ImgBufferOut[OutputBufferCoord] = 0;
+                OutputBufferCoord++;
+                }
+                // move pointer to next pixel
+                CurrCoord += 2 * 3;
+            }
+            #pragma omp atomic
+            iRowCount++;
+            //update progress bar only every N rows
+            if (iRowCount % 200 == 0) {
+                 #pragma omp critical
+                 {
+                 gimp_progress_update ((gdouble) (iRowCount - y1) / (gdouble) (imgheight));
+                 }
             }
-            // move pointer to next pixel
-            CurrCoord += 2 * 3;
         }
-        //update progress bar for every N rows
-        if (i % 20 == 0) gimp_progress_update ((gdouble) (i - y1) / (gdouble) (imgheight));
+        g_free(UndistCoord);
+        lf_free(mod);
     }
   
     //write data back to gimp
@@ -799,7 +824,6 @@ static void process_image (GimpDrawable *drawable) {
     // free memory
     g_free(ImgBufferOut);
     g_free(ImgBuffer);
-    g_free(UndistCoord);
 
     lf_free(lenses);
     lf_free(cameras);
